@@ -746,7 +746,17 @@ regex = /bin/bash"
         let mut vec_eo: Vec<EnvOptions> = vec![];
         let mut bytes = 0;
         read_ini_config_str(&config, &mut vec_eo, "ed", false, &mut bytes);
-        assert_eq!(vec_eo.iter().next().unwrap().rule.as_str(), "/bin/bash");
+        assert_eq!(
+            vec_eo
+                .iter()
+                .next()
+                .unwrap()
+                .rule
+                .as_ref()
+                .unwrap()
+                .as_str(),
+            "/bin/bash"
+        );
 
         let mut ro = RunOptions::new();
         ro.name = "ed".to_string();
@@ -982,7 +992,24 @@ editmode=0644
 
         let entry = can(&vec_eo, &ro);
 
-        assert_eq!(entry.edit_mode, Some(420));
+        assert!(matches!(entry.edit_mode, Some(EditMode::Mode(420))));
+
+        let config = "
+[edit_filemode]
+name=ed
+target=root
+regex=/etc/please.ini.*
+type=edit
+editmode=keep
+"
+        .to_string();
+
+        bytes = 0;
+        read_ini_config_str(&config, &mut vec_eo, "ed", false, &mut bytes);
+
+        let entry = can(&vec_eo, &ro);
+
+        assert!(matches!(entry.edit_mode, Some(EditMode::Keep(true))));
     }
 
     #[test]
@@ -1440,4 +1467,67 @@ regex = /bin/echo hello[\\\\]world
 
         assert_eq!(can(&vec_eo, &ro).permit, true);
     }
+
+    #[test]
+    fn test_replace_new_args_spaces_exact_rule() {
+        let config = r#"
+[ed]
+name = ed
+type = run
+target = root
+exact_rule = /bin/echo hello\ world
+"#
+        .to_string();
+        let mut vec_eo: Vec<EnvOptions> = vec![];
+        let mut bytes = 0;
+        read_ini_config_str(&config, &mut vec_eo, "ed", false, &mut bytes);
+        let mut ro = RunOptions::new();
+        ro.name = "ed".to_string();
+        ro.target = "root".to_string();
+        ro.command = replace_new_args(vec!["/bin/echo".to_string(), "hello world".to_string()]);
+
+        assert_eq!(can(&vec_eo, &ro).permit, true);
+    }
+
+    #[test]
+    fn test_replace_new_args_spaces_rule() {
+        let config = r#"
+[ed]
+name = ed
+type = run
+target = root
+rule = /bin/echo hello\\ world
+"#
+        .to_string();
+        let mut vec_eo: Vec<EnvOptions> = vec![];
+        let mut bytes = 0;
+        read_ini_config_str(&config, &mut vec_eo, "ed", false, &mut bytes);
+        let mut ro = RunOptions::new();
+        ro.name = "ed".to_string();
+        ro.target = "root".to_string();
+        ro.command = replace_new_args(vec!["/bin/echo".to_string(), "hello world".to_string()]);
+
+        assert_eq!(can(&vec_eo, &ro).permit, true);
+    }
+
+    #[test]
+    fn test_replace_new_args_internal_backslash_spaces_rule() {
+        let config = r#"
+[ed]
+name = ed
+type = run
+target = root
+rule = /bin/echo hello\\ \\\\\\ world
+"#
+        .to_string();
+        let mut vec_eo: Vec<EnvOptions> = vec![];
+        let mut bytes = 0;
+        read_ini_config_str(&config, &mut vec_eo, "ed", false, &mut bytes);
+        let mut ro = RunOptions::new();
+        ro.name = "ed".to_string();
+        ro.target = "root".to_string();
+        ro.command = replace_new_args(vec!["/bin/echo".to_string(), "hello \\ world".to_string()]);
+
+        assert_eq!(can(&vec_eo, &ro).permit, true);
+    }
 }
